
#define BANNER "@(#)RD58DBG LOADER 0.0 970207"

#define MKWORD(lo,hi) ((lo) | ((hi) << 8))

#define BUF(n) : .rs n
#define BYTE   : .rs 1
#define WORD   : .rs 2

!----------------------------------------------------------------------
!
!	I/O map
!
#define PIO   0x00
#define SIO   0x10
#define TMR   0x20
#define DA0   0x30
#define DA1   0x40
#define AD    0x50
#define OUT0  0x60
#define OUT1  0x70
#define OUT2  0x80
#define WD    0x90
#define MODEM 0xA0

#define ADATA 0
#define BDATA 1
#define ACTRL 2
#define BCTRL 3

#define TMRCTRL 3

!----------------------------------------------------------------------
!
!	8254 timer
!

#define TMR_0		0x00
#define TMR_1		0x40
#define TMR_2		0x80
#define TMR_INTTC	0x00	/* mode 0, intr on terminal cnt */
#define TMR_ONESHOT	0x02	/* mode 1, one shot */
#define TMR_RATEGEN	0x04	/* mode 2, rate generator */
#define TMR_SQWAVE	0x06	/* mode 3, square wave */
#define TMR_SWTRIGG	0x08	/* mode 4, s/w triggered strobe */
#define TMR_LSB		0x10	/* r/w counter LSB */
#define TMR_MSB		0x20	/* r/w counter MSB */
#define TMR_BOTH	0x30	/* r/w counter 16 bits, LSB first */

!----------------------------------------------------------------------
!
!	SIO defs
!

#define WR0_SEND_ABORT			0x08
#define WR0_RESET_ESCINT		0x10
#define WR0_CHANNEL_RESET		0x18
#define WR0_REARM_1STC			0x20
#define WR0_RESET_TXINT			0x28
#define WR0_ERROR_RESET			0x30
#define WR0_RETINT				0x38
#define WR0_RESET_RXCRC			0x40
#define WR0_RESET_TXCRC			0x80
#define WR0_RESET_UNDERRUN		0xc0

#define WR1_ESCINT_ENB			0x01
#define WR1_TXINT_ENB			0x02
#define WR1_STATUS_VECTOR		0x04
#define WR1_RXINT_DISABLE		0x00
#define WR1_INT_1STC			0x08
#define WR1_INT_ALLC			0x10
#define WR1_INT_ALLC_PNS		0x18
#define WR1_WAITREADY_RT		0x20
#define WR1_WAITREADY_FN		0x40
#define WR1_WAITREADY_ENB		0x80

#define WR3_RX_ENB				0x01
#define WR3_SYNCLOAD_INHIBIT	0x02
#define WR3_ADDRESS_SEARCH_MODE	0x04
#define WR3_RXCRC_ENB			0x08
#define WR3_ENTER_HUNT			0x10
#define WR3_AUTO_ENABLES		0x20
#define WR3_5BIT_RX				0x00
#define WR3_6BIT_RX				0x40
#define WR3_7BIT_RX				0x80
#define WR3_8BIT_RX				0xc0

#define WR4_PARITY_ENB			0x01
#define WR4_PARITY_ODD			0x00
#define WR4_PARITY_EVEN			0x02
#define WR4_SYNC_MODES			0x00
#define WR4_1STOPBIT			0x04
#define WR4_15STOPBIT			0x08
#define WR4_2STOPBIT			0x0c
#define WR4_8BITSYNC			0x00
#define WR4_16BITSYNC			0x10
#define WR4_SDLC_MODE			0x20
#define WR4_EXTSYNC				0x30
#define WR4_X1_CLK				0x00	/* 153600 bd, but... */
#define WR4_X16_CLK				0x40	/*   9600 bd */
#define WR4_X32_CLK				0x80	/*   4800 bd */
#define WR4_X64_CLK				0xc0	/*   2400 bd */

#define WR5_TXCRC_ENB			0x01
#define WR5_RTS					0x02
#define WR5_CCITT_CRC			0x00
#define WR5_CRC16_CRC			0x04
#define WR5_TX_ENB				0x08
#define WR5_SEND_BREAK			0x10
#define WR5_5BIT_TX				0x00
#define WR5_7BIT_TX				0x20
#define WR5_6BIT_TX				0x40
#define WR5_8BIT_TX				0x60
#define WR5_DTR					0x80

#define RR0_RCA					0x01
#define RR0_INT_PENDING			0x02
#define RR0_TBE					0x04
#define RR0_DCD					0x08
#define RR0_SYNC				0x10
#define RR0_CTS					0x20
#define RR0_UNDERRUN			0x40
#define RR0_BREAK_ABORT			0x80

#define RR1_ALL_SENT			0x01

#define RR1_PARER				0x10
#define RR1_ROVER				0x20
#define RR1_FRMERR				0x40
#define RR1_CRCERR				0x40
#define RR1_ENDFRAME			0x80

!----------------------------------------------------------------------
!
!	Output latch bits
!
#define O0_VOLUME 0x07
#define O0_INH    0x08
#define O0_AUDIOC 0x10
#define O0_CCIRC  0x20
#define O0_MTC    0x40
#define O0_MICM   0x80

#define O1_SRE    0x01
#define O1_SCE    0x02
#define O1_STE    0x04
#define O1_CLK    0x08
#define O1_SD     0x10
#define O1_RAS    0x20
#define O1_TPS    0x40
#define O1_TXOFF  0x80

!
!	Changes only by mainline.
!

#define O2_RA14   0x01
#define O2_RA15   0x02
#define O2_RS     0x04
#define O2_SMEM   0x08
#define O2_CS1    0x10
#define O2_CS2    0x20
#define O2_CLK    0x40
#define O2_DP     0x80

#define O2_XXX    O2_SMEM

#define O2_KEYPAD (O2_XXX | 0)
#define O2_LATCH  (O2_XXX | O2_CS1)
#define O2_LCD1   (O2_XXX | O2_CS2)
#define O2_LCD2   (O2_XXX | O2_CS2 | O2_CS1)

!----------------------------------------------------------------------
!
!	PIO and SIO controls
!
#define PA_CLK2   0x01
#define PA_HOOK   0x02
#define PA_WDR    0x04
#define PA_PWR    0x08
#define PA_CCIR   0xf0	/* all inputs */

#define PB_RAMA12 0x01	/* out */
#define PB_EXIN1  0x02
#define PB_EXIN2  0x04
#define PB_DCU    0x08
#define PB_RXOFF  0x10
#define PB_TMR0   0x20
#define PB_EXAL   0x40	/* out */
#define PB_PWROFF 0x80	/* out; 3eh are inputs */

#define SA_KKINT  RR0_DCD
#define SA_DA     RR0_CTS

#define SB_PTT    RR0_CTS
#define SB_LOCAL  RR0_SYNC
#define SB_DMIDLE RR0_DCD	/* No MBUS activity */
#define SB_MON    WR5_DTR	/* Master ON pulse */

!	LATCH bits

#define IND_CALL     0x02
#define IND_ROAM     0x04
#define IND_KEYLIGHT 0x08
#define IND_SERV     0x10
#define IND_ON       0x20
#define IND_LCDLIGHT 0x40
#define IND_BRIGHT   0x80

!	Special LCD segments

#define SEG_V_U      0x43
#define SEG_V_D      0x53
#define SEG_PHONE    0x47
#define SEG_CLOCK    0x4b
#define SEG_COLON_UR 0x4f
#define SEG_COLON_UL 0x63
#define SEG_COLON_D  0x57
#define SEG_PHONE_NO 0x5b
#define SEG_MAST     0x5f
#define SEG_BAR_L    0x67
#define SEG_BAR_R    0x6f
#define SEG_STAR     0x6b
#define SEG_KEY      0x73
#define SEG_BOOK     0x77
#define SEG_CAR_D    0x7b
#define SEG_CAR_U    0x7f

!----------------------------------------------------------------------
!
!
!
!	N = (Frx + 86512.5) / 1600
!	A = (Frx + 86512.5) / 12.5 - N * 128
!
!	N = (433500 + 86512.5) / 1600
!	N = 325
!
!	A = (433500 + 86512.5) / 12.5 - 325 * 128
!	A = 1
!

RXOFFSET = (9 | (54 << 7))		! 54,9 for 1st IF (86512.5)

	.data
	.org 0xc000

output_0	:.rs 1
output_1	:.rs 1
output_2	:.rs 1

mbusrx_rp	:.rs 2
mbusrx_wp	:.rs 2

mbustx_rp	:.rs 2
mbustx_wp	:.rs 2

	.align 8

_bss:

mbusrx_buf	:.rs 256		! page aligned
mbustx_buf	:.rs 256		! page aligned

mbustx_cnt	:.rs 1
mbusrx_cnt	:.rs 1

mbustxon	:.rs 1		! timer for MBUS transmit

mt_timer	:.rs 1

hack		:.rs 1

_end:

!----------------------------------------------------------------------
!
!	RESET
!
	.text
	.org 0

	di
	out [WD], a
	jp start

banner:
	.asciz BANNER

!----------------------------------------------------------------------
!
!	NMI
!
	.org 0x0066

	out [WD], a

	ld sp, nmiretstk	! restart
	retn
nmiretstk:
	.word 0				! to beginning

!----------------------------------------------------------------------
!
!	Interrupt table, short-aligned, on one page.
!
	.org 0x0080

intvec:

sio_base:
	.word siob_tbe
	.word siob_esc
	.word siob_rca
	.word siob_src
	.word sioa_tbe
	.word sioa_esc
	.word sioa_rca
	.word sioa_src

pio_base:
	.word pioa_int
	.word piob_int

!----------------------------------------------------------------------
!
!	Initialization commands for PIO, SIO and 8254
!	In that order, A/B, A/B, 0/1/2 !
!

init_chips:
1:
	.byte LO(pio_base)	! load interrupt vector
	.byte 0xcf			! set operating mode 3
	.byte 0xff			! i/o selection, 1 for input, 0 for output
	.byte 0x97			! set interrupt control, IE, OR, LOW, MASK
	.byte 0xff			! input bit monitoring mask, none.
PIOA_INIT = . - 1b

1:
	.byte LO(pio_base+2)	! load interrupt vector
	.byte 0xcf			! set operating mode 3
	.byte 0x3e			! i/o selection, 1 for input, 0 for output
	.byte 0x97			! set interrupt control, IE, OR, LOW, MASK
	.byte 0xdf			! input bit monitoring mask, only B5 (TMR0)
PIOB_INIT = . - 1b

1:
	.byte WR0_CHANNEL_RESET
	.byte 4, WR4_1STOPBIT | WR4_X16_CLK
	.byte 3, WR3_8BIT_RX
	.byte 5, WR5_8BIT_TX | WR5_RTS | WR5_DTR
	.byte 1, WR1_STATUS_VECTOR | WR1_ESCINT_ENB
SIOA_INIT = . - 1b

1:
	.byte WR0_CHANNEL_RESET
	.byte 4, WR4_1STOPBIT | WR4_X16_CLK
	.byte 3, WR3_8BIT_RX | WR3_RX_ENB
	.byte 5, WR5_8BIT_TX | WR5_RTS | WR5_DTR | WR5_TX_ENB
	.byte 1, WR1_STATUS_VECTOR | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC
	.byte 2, LO(sio_base)
SIOB_INIT = . - 1b

	!  Control words, then initial counts.
1:
	.byte TMR_0 | TMR_BOTH | TMR_SQWAVE
	.byte TMR_1 | TMR_BOTH | TMR_SQWAVE
	.byte TMR_2 | TMR_BOTH | TMR_RATEGEN
TMR_INIT = . - 1b

	.word 4		! LP Filter
	.word 4		! TX- and audible tones
	.word 20		! 100 Hz clock interrupt to PIO B5

!
!	All timer-gates are fixed on.
!	timer-clocks 0 and 1 are 4.032 MHz, 2 is 1968.75 Hz
!
!	System hz is 100, from PIO B5 and TMR0_2
!
MT_600HZ  equ 4032000 / 600

!----------------------------------------------------------------------
!
!	Initialization before main
!

start:

	ld sp, 0			! stack down from end of ram (last 16kB)

	!
	!  Output latches
	!

	ld a, 0				! volume setting at lowest level
	ld [output_0], a
	out [OUT0], a

	ld a, O1_TXOFF		! no tx, idle all
	ld [output_1], a
	out [OUT1], a

	ld a, O2_SMEM		! select ram 0, i hope
	ld [output_2], a
	out [OUT2], a

	out [WD], a

	!
	!  PIO, SIO and 8254
	!

	ld hl, init_chips

	ld bc, MKWORD(PIO + ACTRL, PIOA_INIT)
	otir
	out [WD], a

	ld bc, MKWORD(PIO + BCTRL, PIOB_INIT)
	otir
	out [WD], a

	! Set output bits, and check if restarted from powerdown nmi.

	ld a, 0
	out [PIO+BDATA], a	! B0, EXAL, OFF all zero

	in a, [PIO+ADATA]
	and PA_PWR
	jp nz, powerdown

	ld bc, MKWORD(SIO + ACTRL, SIOA_INIT)
	in a, [c]
	otir
	out [WD], a

	ld bc, MKWORD(SIO + BCTRL, SIOB_INIT)
	in a, [c]
	otir
	out [WD], a

	ld bc, MKWORD(TMR + TMRCTRL, TMR_INIT)
	otir
	ld bc, MKWORD(TMR + 0, 2)
	otir
	ld bc, MKWORD(TMR + 1, 2)
	otir
	ld bc, MKWORD(TMR + 2, 2)
	otir

	out [WD], a

	ld a, HI(intvec)
	ld iv, a
	im2

	ld hl, mbusrx_buf
	ld [mbusrx_rp], hl
	ld [mbusrx_wp], hl

	ld hl, mbustx_buf
	ld [mbustx_rp], hl
	ld [mbustx_wp], hl

	out [WD], a

	ld hl, _bss					! 1st to zero
	ld de, _bss + 1				! 2nd to zero
	ld bc, _end - _bss - 1		! loop count
	ld [hl], 0					! zero 1st
	ldir						! and rest

	out [WD], a

	ld a, LO(MT_600HZ)
	out [TMR + 1], a
	ld a, HI(MT_600HZ)
	out [TMR + 1], a
	ld a, [output_0]
	or O0_MTC
	ld [output_0], a
	out [OUT0], a
	ld a, 5
	ld [mt_timer], a

	ei

	jp main

!----------------------------------------------------------------------

doreti:
	pop af
	ei
	reti

!----------------------------------------------------------------------
!
!  SIO interrupts
!

siob_tbe:
	push af
	ld a, [mbustx_cnt]
	dec a
	ld [mbustx_cnt], a
	jr z, siob_tbe_close
	push hl
	ld hl, [mbustx_rp]
	ld a, [hl]
	inc l
	ld [mbustx_rp], hl
	pop hl
	out [SIO+BDATA], a
	ld a, 25
	ld [mbustxon], a		! keep transmit timer on
	jp doreti

siob_tbe_close:
	ld a, WR0_RESET_TXINT
	out [SIO+BCTRL], a
	jp doreti

siob_rca_drop:
	ld a, LO(MT_600HZ)
	out [TMR + 1], a
	ld a, HI(MT_600HZ)
	out [TMR + 1], a
	ld a, [output_0]
	or O0_MTC
	ld [output_0], a
	out [OUT0], a
	ld a, 100
	ld [mt_timer], a

siob_rca_ignore:
	in a, [SIO+BDATA]

	jp doreti

siob_rca:
	push af
	ld a, [mbustxon]
	or a
	jr nz, siob_rca_ignore
	ld a, [mbusrx_cnt]
	inc a
	jr z, siob_rca_drop			! no room
	ld [mbusrx_cnt], a
	push hl
	ld hl, [mbusrx_wp]
	in a, [SIO+BDATA]
	ld [hl], a
	inc l
	ld [mbusrx_wp], hl
	pop hl
	jp doreti

!
!	SIO B ext/status change
!
siob_esc:
	push af
	ld a, WR0_RESET_ESCINT
	out [SIO+BCTRL], a
	jp doreti

!
!	SIO B special receive condition
!
siob_src:
	push af
	in a, [SIO+BDATA]
	ld a, WR0_ERROR_RESET
	out [SIO+BCTRL], a

	ld a, LO(MT_600HZ)
	out [TMR + 1], a
	ld a, HI(MT_600HZ)
	out [TMR + 1], a
	ld a, [output_0]
	or O0_MTC
	ld [output_0], a
	out [OUT0], a
	ld a, 100
	ld [mt_timer], a

	jp doreti

sioa_tbe:
	push af
	ld a, WR0_RESET_TXINT
	out [SIO+ACTRL], a
	jp doreti

sioa_rca:
	push af
	in a, [SIO+ADATA]
	jp doreti

!
!	SIO A ext/status change
!
sioa_esc:
	push af
#if 1
	in a, [SIO+ACTRL]
	and SA_DA
	jp nz, sioa_esc_1
	ld a, 1
	ld [hack], a
sioa_esc_1:
#endif
	ld a, WR0_RESET_ESCINT
	out [SIO+ACTRL], a
	jp doreti

!
!	SIO A special receive condition
!
sioa_src:
	push af
	in a, [SIO+ADATA]
	ld a, WR0_ERROR_RESET
	out [SIO+ACTRL], a
	jp doreti

!----------------------------------------------------------------------
!
!  PIO interrupts
!
pioa_int:
	push af
	jp doreti

piob_int:
	push af
	out [WD], a				! keep watchdog happy.
	in a, [TMR+TMRCTRL]		! dummy read to raise PIO B5 (remove int)

	ld a, [mt_timer]
	or a
	jr z, piob_int_1
	dec a
	ld [mt_timer], a
	jr nz, piob_int_1
	ld a, [output_0]
	and ~O0_MTC
	ld [output_0], a
	out [OUT0], a

piob_int_1:

	ld a, [mbustxon]
	or a
	jr z, piob_int_2
	dec a
	ld [mbustxon], a

piob_int_2:

	jp doreti

!----------------------------------------------------------------------
!
!	Mainline
!

main:

	ld c, 0xa
	call putchar

	ld de, banner
1:
	ld a, [de]
	or a
	jr z, 1f
	ld c, a
	call putchar
	inc de
	jr 1b
1:
	ld c, 0xa
	call putchar

	ld a, 0
	ld [hack], a

main_1:

	call getchar
	cp 0xf3
	jr nz, main_1

	ld de, 0xd000

main_2:

	ld [de], a
	inc de

	call getchar

	jp main_2

showit:

	ld de, 0x8000

showit_1:

	ld a, [de]
	ld c, a
	call putchar
	
	inc de
	ld a, d
	or e
	jr nz, showit_1

	jp main

!----------------------------------------------------------------------
!
!	Powerdown
!
powerdown:
	out [WD], a
	ld a, PB_PWROFF
	out [PIO+BDATA], a
	halt
	jr powerdown

!----------------------------------------------------------------------
!
!	MBUS putchar, character in c
!

putchar:
	ld hl, mbustx_cnt
putchar_1:
	ld a, [hl]
	inc a
	jr z, putchar_1			! wait until not full
	di
	ld a, [hl]
	inc [hl]
	or a
	jr z, putchar_start		! tx not active, go start it
	ld hl, [mbustx_wp]
	ld [hl], c
	inc l
	ld [mbustx_wp], hl
	ei
	ret
putchar_start:
	ld a, c
	out [SIO+BDATA], a
	ld a, 25
	ld [mbustxon], a
putchar_out:
	ei
	ret

getchar:
	ld a, [hack]
	or a
	jr z, getchar_1
	pop hl
	ld hl, 0xd000
	push hl
	di
	ret						! go to ram

getchar_1:
	ld hl, mbusrx_cnt
	ld a, [hl]
	or a
	jr z, getchar
	di
	dec [hl]
	ld hl, [mbusrx_rp]
	ld a, [hl]
	inc l
	ld [mbusrx_rp], hl
	ei
	ret

